# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 15:31:13 2024

@author: mstep
"""

#!/usr/bin/env python
# coding: utf-8

#%%

def rgb_to_hex(r, g, b):
    '''Convert RGB values into a hex string that tkinter can understand.'''
    r = int(r)
    g = int(g)
    b = int(b)
    return f'#{r:02x}{g:02x}{b:02x}'

#%%

import math

def get_min_max_corners(points):
    '''
    Find the points nearest the upper left, upper right,
    lower left, and lower right corners.
    '''
    # Start with the first point as the solution.
    ul = points[0]
    ur = ul
    ll = ul
    lr = ul

    # Search the other points.
    for point in points:
        if -point[0] - point[1] > -ul[0] - ul[1]:   ul = point
        if  point[0] - point[1] >  ur[0] - ur[1]:   ur = point
        if -point[0] + point[1] > -ll[0] + ll[1]:   ll = point
        if  point[0] + point[1] >  lr[0] + lr[1]:   lr = point
    return ul, ur, ll, lr

def get_min_max_box(points):
    '''Find a box that fits inside the MinMax quadrilateral.'''
    # Find the MinMax quadrilateral.
    ul, ur, ll, lr = get_min_max_corners(points)
    quad = [ul, ur, lr, ll]

    # Get the coordinates of a box that lies inside this quadrilateral.
    xmin = ul[0]
    ymin = ul[1]

    xmax = ur[0]
    if (ymin < ur[1]): ymin = ur[1]

    if (xmax > lr[0]): xmax = lr[0]
    ymax = lr[1]

    if (xmin < ll[0]): xmin = ll[0]
    if (ymax > ll[1]): ymax = ll[1]

    box = [xmin, xmax, ymin, ymax]

    return quad, box

def hull_cull(points):
    '''
    Cull points out of the convex hull that lie inside the
    quadezoid defined by the vertices with smallest and
    largest X and Y coordinates.
    '''
    # Find a culling box.
    quad, box = get_min_max_box(points)
    xmin, xmax, ymin, ymax = box
    rect = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]

    # Cull the points.
    not_culled = []
    culled = []
    for point in points:
        # See if (this point lies outside of the culling box.
        if (point[0] <= xmin or point[0] >= xmax or \
            point[1] <= ymin or point[1] >= ymax):
            # This point cannot be culled.
            not_culled.append(point)
        else:
            # This point is culled.
            culled.append(point)

    return quad, rect, culled, not_culled

def find_convex_hull(points):
    '''Find the points' convex hull.'''
    # Make a copy so we don't messs up the points list.
    points = points.copy()

    # Cull.
    quad, rect, culled, not_culled = hull_cull(points)

    # Find the remaining point with the smallest Y value.
    # If there's a tie, take the one with the smaller X value.
    best_point = points[0]
    for point in points:
        if ((point[1] < best_point[1]) or
                ((point[1] == best_point[1]) and (point[0] < best_point[0]))):
            best_point = point

    # Add this point to the convex hull.
    hull = [best_point]
    points.remove(best_point)

    # Start wrapping up the other points.
    sweep_angle = 0
    while len(points) > 0:
        # Find the point with smallest angle_value from the last point.
        x = hull[-1][0]
        y = hull[-1][1]
        best_point = points[0]
        best_angle = 3600

        # Search the rest of the points.
        for point in points:
            test_angle = angle_value(x, y, point[0], point[1])
            if (test_angle >= sweep_angle) and (best_angle > test_angle):
                best_angle = test_angle
                best_point = point

        # See if the first point is better.
        # If so, we are done.
        first_angle = angle_value(x, y, hull[0][0], hull[0][1])
        if (first_angle >= sweep_angle) and (best_angle >= first_angle):
            # The first point is better. We're done.
            break

        # Add the best point to the convex hull.
        hull.append(best_point)
        points.remove(best_point)

        sweep_angle = best_angle

    return quad, rect, culled, not_culled, hull

def angle_value(x1, y1, x2, y2):
    '''
    Return a number that gives the ordering of angles WRST horizontal from the
    point (x1, y1) to (x2, y2). In other words, angle_value(x1, y1, x2, y2) is
    not the angle, but if:
        angle_value(x1, y1, x2, y2) > angle_value(x1, y1, x2, y2)
    then
        angle(x1, y1, x2, y2) > angle(x1, y1, x2, y2)
    This function is dy / (dy + dx).
    '''
    dx = x2 - x1
    ax = abs(dx)
    dy = y2 - y1
    ay = abs(dy)
    if math.isclose(ax + ay, 0):
        # If (the two points are the same, return 360.
        t = 360 / 9
    else:
        t = dy / (ax + ay)

    if dx < 0:
        t = 2 - t
    elif dy < 0:
        t = 4 + t
    return t * 90

def find_convex_hulls(points):
    '''Extract convex hulls from the points until there's nothing left.'''
    # Make a copy so we don't messs up the points list.
    points = points.copy()

    hulls = []
    while len(points) > 0:
        # Find the convex hull.
        _, _, _, _, hull = find_convex_hull(points)

        # Save the hull.
        hulls.append(hull)

        # Remove this hull's points from the points list.
        for point in hull:
            points.remove(point)

    return hulls

# In[2]:

import random
import tkinter as tk

class ConvexHullsColoredApp:
    def __init__(self):
        # Start with no hulls.
        self.hulls = []

        # Initialize tkinter.
        self.window = tk.Tk()
        self.window.title('convex_hulls_colored')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('600x400')

        # Left frame.
        frame = tk.Frame(self.window)
        frame.pack(side=tk.LEFT, anchor=tk.N, padx=3, pady=3)

        # # Points.
        LABEL_WIDTH = 14
        ENTRY_WIDTH = 8
        self.num_points_entry, self.num_points_var = \
            self.make_field(frame, LABEL_WIDTH, "# Points:",
                            ENTRY_WIDTH, '100')

        # Buttons.
        BUTTON_WIDTH = 15
        self.generate_button = self.make_button(
            frame, BUTTON_WIDTH, 'Generate', self.generate_click, tk.NORMAL)

        self.find_hull_button = self.make_button(
            frame, BUTTON_WIDTH, 'Find Hulls', self.find_hulls_click,
            tk.DISABLED)

        # Canvas.
        self.canvas = tk.Canvas(self.window, bg='white', borderwidth=2,
                                relief=tk.SUNKEN)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH,
                         expand=True, padx=(0,5), pady=5)

        # Bind the Return key.
        self.window.bind("<Return>", lambda e: self.generate_click())

        self.num_points_entry.focus_force()
        self.window.mainloop()

    def make_field(self, parent, label_width, label_text,
                   entry_width, entry_text, entry_state=tk.NORMAL):
        '''
        Make a Label and associated Entry in columns 0 and 1.
        Return the Entry and its variable.
        '''
        frame = tk.Frame(parent)
        frame.pack(side=tk.TOP)

        label = tk.Label(frame, text=label_text, width=label_width,
            anchor=tk.W)
        label.pack(side=tk.LEFT, padx=3, pady=3)

        var = tk.StringVar(parent, value=entry_text)
        entry = tk.Entry(frame, textvariable=var, width=entry_width,
            justify=tk.RIGHT, state=entry_state)
        entry.pack(side=tk.LEFT, padx=3, pady=5)
        return entry, var

    def make_button(self, parent, button_width, text, command, state):
        '''Make a Button in this row spanning columns 0 and 1.'''
        button = tk.Button(parent, text=text, width=button_width,
                           command=command, state=state)
        button.pack(side=tk.TOP, padx=3, pady=3)
        return button

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def generate_click(self):
        # Pick random points.
        wid = self.canvas.winfo_width()
        hgt = self.canvas.winfo_height()
        MARGIN = 20
        self.num_points = int(self.num_points_var.get())
        self.points = []
        for _ in range(self.num_points):
            x = random.randint(MARGIN, wid - MARGIN)
            y = random.randint(MARGIN, hgt - MARGIN)
            self.points.append((x, y))

        # Enable the Find Hull button.
        self.find_hull_button.configure(state=tk.NORMAL)

        # We have no hull.
        self.hulls = []

        # Draw the points.
        self.draw()

    def find_hulls_click(self):
        '''Solve by using random orderings of the points.'''
        self.hulls = find_convex_hulls(self.points)

        # Disable the Find Hull button.
        self.find_hull_button.configure(state=tk.DISABLED)

        # Draw the results.
        self.draw()

    def draw_point(self, point, color):
        RADIUS = 5
        RADIUS = 1
        self.canvas.create_oval(
            point[0] - RADIUS, point[1] - RADIUS,
            point[0] + RADIUS, point[1] + RADIUS,
            fill=color)

    def draw(self):
        '''Draw the points, hull, etc.'''
        # Delete any previous drawing objects.
        self.canvas.delete(tk.ALL)

        # Calculate colors.
        dc = 255 / (len(self.hulls) - 1)

        # Draw the hulls.
        r = 255
        g = 0
        b = 0
        outline = ''
        for hull in self.hulls:
            self.canvas.create_polygon(hull, outline=outline,
                                       fill=rgb_to_hex(r, g, b))
            g += dc
            b += dc

        # Draw the points.
        for point in self.points:
            self.draw_point(point, 'blue')

# In[3]:

ConvexHullsColoredApp()
print('Done')


# In[ ]:
